from sqlalchemy import text, types, util
from sqlalchemy.engine import default
from sqlalchemy.sql import compiler

import pydruid.db

RESERVED_SCHEMAS = ["INFORMATION_SCHEMA"]


jdbc_type_map = {
    -6: types.BigInteger,
    -5: types.BigInteger,
    1: types.String,
    3: types.Float,
    4: types.BigInteger,
    5: types.BigInteger,
    6: types.Float,
    7: types.Float,
    8: types.Float,
    12: types.String,
    16: types.Boolean,
    91: types.DATE,
    93: types.TIMESTAMP,
    1111: types.BLOB,
}


class UniversalSet(object):
    def __contains__(self, item):
        return True


class DruidIdentifierPreparer(compiler.IdentifierPreparer):
    reserved_words = UniversalSet()


class DruidCompiler(compiler.SQLCompiler):
    pass


class DruidTypeCompiler(compiler.GenericTypeCompiler):
    def visit_REAL(self, type_, **kwargs):
        return "DOUBLE"

    def visit_NUMERIC(self, type_, **kwargs):
        return "LONG"

    visit_DECIMAL = visit_NUMERIC
    visit_INTEGER = visit_NUMERIC
    visit_SMALLINT = visit_NUMERIC
    visit_BIGINT = visit_NUMERIC
    visit_BOOLEAN = visit_NUMERIC
    visit_TIMESTAMP = visit_NUMERIC
    visit_DATE = visit_NUMERIC

    def visit_CHAR(self, type_, **kwargs):
        return "STRING"

    visit_NCHAR = visit_CHAR
    visit_VARCHAR = visit_CHAR
    visit_NVARCHAR = visit_CHAR
    visit_TEXT = visit_CHAR

    def visit_DATETIME(self, type_, **kwargs):
        return "LONG"

    def visit_TIME(self, type_, **kwargs):
        return "LONG"

    def visit_BLOB(self, type_, **kwargs):
        return "COMPLEX"

    visit_CLOB = visit_BLOB
    visit_NCLOB = visit_BLOB
    visit_VARBINARY = visit_BLOB
    visit_BINARY = visit_BLOB


class DruidDialect(default.DefaultDialect):

    name = "druid"
    scheme = "http"
    driver = "rest"
    user = None
    password = None
    preparer = DruidIdentifierPreparer
    statement_compiler = DruidCompiler
    type_compiler = DruidTypeCompiler
    supports_alter = False
    supports_pk_autoincrement = False
    supports_default_values = False
    supports_empty_insert = False
    supports_unicode_statements = True
    supports_unicode_binds = True
    returns_unicode_strings = True
    description_encoding = None
    supports_native_boolean = True

    def __init__(self, context=None, *args, **kwargs):
        super(DruidDialect, self).__init__(*args, **kwargs)
        self.context = context or {}

    @classmethod
    def dbapi(cls):
        return pydruid.db

    def create_connect_args(self, url):
        kwargs = {
            "host": url.host,
            "port": url.port or 8082,
            "user": url.username or None,
            "password": url.password or None,
            "path": url.database,
            "scheme": self.scheme,
            "context": self.context,
            "header": url.query.get("header") == "true",
        }
        return ([], kwargs)

    def do_ping(self, dbapi_connection) -> bool:
        """
        Return if the database can be reached.
        """
        try:
            dbapi_connection.execute("SELECT 1")
        except Exception:
            return False

        return True

    def get_schema_names(self, connection, **kwargs):
        # Each Druid datasource appears as a table in the "druid" schema. This
        # is also the default schema, so Druid datasources can be referenced as
        # either druid.dataSourceName or simply dataSourceName.
        result = connection.execute(
            text("SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA")
        )

        return [
            row.SCHEMA_NAME for row in result if row.SCHEMA_NAME not in RESERVED_SCHEMAS
        ]

    def has_table(self, connection, table_name, schema=None):
        query = """
            SELECT COUNT(*) > 0 AS exists_
              FROM INFORMATION_SCHEMA.TABLES
             WHERE TABLE_NAME = '{table_name}'
        """.format(
            table_name=table_name
        )

        result = connection.execute(text(query))
        return result.fetchone().exists_

    def get_table_names(self, connection, schema=None, **kwargs):
        query = "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES"
        if schema:
            query = "{query} WHERE TABLE_SCHEMA = '{schema}'".format(
                query=query, schema=schema
            )

        result = connection.execute(text(query))
        return [row.TABLE_NAME for row in result]

    def get_view_names(self, connection, schema=None, **kwargs):
        return []

    def get_table_options(self, connection, table_name, schema=None, **kwargs):
        return {}

    def get_columns(self, connection, table_name, schema=None, **kwargs):
        query = """
            SELECT COLUMN_NAME,
                   JDBC_TYPE,
                   IS_NULLABLE,
                   COLUMN_DEFAULT
              FROM INFORMATION_SCHEMA.COLUMNS
             WHERE TABLE_NAME = '{table_name}'
        """.format(
            table_name=table_name
        )
        if schema:
            query = "{query} AND TABLE_SCHEMA = '{schema}'".format(
                query=query, schema=schema
            )

        result = connection.execute(text(query))

        return [
            {
                "name": row.COLUMN_NAME,
                "type": self._map_jdbc_type(row),
                "nullable": get_is_nullable(row.IS_NULLABLE),
                "default": get_default(row.COLUMN_DEFAULT),
            }
            for row in result
        ]

    def get_pk_constraint(self, connection, table_name, schema=None, **kwargs):
        return {"constrained_columns": [], "name": None}

    def get_foreign_keys(self, connection, table_name, schema=None, **kwargs):
        return []

    def get_check_constraints(self, connection, table_name, schema=None, **kwargs):
        return []

    def get_table_comment(self, connection, table_name, schema=None, **kwargs):
        return {"text": ""}

    def get_indexes(self, connection, table_name, schema=None, **kwargs):
        return []

    def get_unique_constraints(self, connection, table_name, schema=None, **kwargs):
        return []

    def get_view_definition(self, connection, view_name, schema=None, **kwargs):
        pass

    def do_rollback(self, dbapi_connection):
        pass

    def _check_unicode_returns(self, connection, additional_tests=None):
        return True

    def _check_unicode_description(self, connection):
        return True

    def _map_jdbc_type(self, row):
        if row.JDBC_TYPE in jdbc_type_map:
            return jdbc_type_map[row.JDBC_TYPE]
        util.warn(
            "Failed to map column '{row.COLUMN_NAME}' with "
            "JDBC type '{row.JDBC_TYPE}' to a sqlalchemy type.".format(row=row)
        )
        return types.NullType


DruidHTTPDialect = DruidDialect


class DruidHTTPSDialect(DruidDialect):

    scheme = "https"


def get_is_nullable(druid_is_nullable):
    # this should be 'YES' or 'NO'; we default to no
    return druid_is_nullable.lower() == "yes"


def get_default(druid_column_default):
    # currently unused, returns ''
    return str(druid_column_default) if druid_column_default != "" else None



from pydruid.utils.dimensions import (
    build_dimension,
    DimensionSpec,
    JavascriptExtraction,
    ListFilteredSpec,
    MapLookupExtraction,
    NamespaceLookupExtraction,
    PartialExtraction,
    RegexExtraction,
    RegexFilteredSpec,
    RegisteredLookupExtraction,
    TimeFormatExtraction,
)


class TestDimensionSpec(object):
    def test_default(self):
        dim_spec = DimensionSpec("dim", "out")
        actual = dim_spec.build()
        expected = {"type": "default", "dimension": "dim", "outputName": "out"}

        assert actual == expected

    def test_extraction_functions(self):
        js_func = "function(x) {return x};"
        ext_fns = [
            (RegexExtraction(r"\w+"), {"type": "regex", "expr": "\\w+"}),
            (PartialExtraction(r"\w+"), {"type": "partial", "expr": "\\w+"}),
            (
                JavascriptExtraction(js_func),
                {"type": "javascript", "function": js_func, "injective": False},
            ),
            (
                MapLookupExtraction(TestMapLookupExtraction.mapping),
                {
                    "type": "lookup",
                    "lookup": {"type": "map", "map": TestMapLookupExtraction.mapping},
                    "retainMissingValue": False,
                    "replaceMissingValueWith": None,
                    "injective": False,
                },
            ),
        ]

        for ext_fn, expected_ext_fn in ext_fns:
            dim_spec = DimensionSpec("dim", "out", extraction_function=ext_fn)
            actual = dim_spec.build()
            expected = {
                "type": "extraction",
                "dimension": "dim",
                "outputName": "out",
                "extractionFn": expected_ext_fn,
            }

            assert actual == expected

    def test_filter_specs(self):
        delegate_spec = DimensionSpec("dim", "out").build()
        filter_specs = [
            (
                ListFilteredSpec(["val1", "val2"]),
                {
                    "type": "listFiltered",
                    "delegate": delegate_spec,
                    "values": ["val1", "val2"],
                },
            ),
            (
                ListFilteredSpec(["val1", "val2"], is_whitelist=False),
                {
                    "type": "listFiltered",
                    "delegate": delegate_spec,
                    "values": ["val1", "val2"],
                    "isWhitelist": False,
                },
            ),
            (
                RegexFilteredSpec(r"\w+"),
                {"type": "regexFiltered", "delegate": delegate_spec, "pattern": "\\w+"},
            ),
        ]

        for filter_spec, expected_dim_spec in filter_specs:
            dim_spec = DimensionSpec("dim", "out", filter_spec=filter_spec)
            actual = dim_spec.build()

            assert actual == expected_dim_spec

    def test_build_dimension(self):
        assert build_dimension("raw_dim") == "raw_dim"

        dim_spec = DimensionSpec("dim", "out")
        assert build_dimension(dim_spec) == dim_spec.build()


class TestListFilteredSpec(object):
    def test_list_filtered_spec(self):
        dim_spec = DimensionSpec("dim", "out").build()
        list_filtered_spec = ListFilteredSpec(["val1", "val2"])
        actual = list_filtered_spec.build(dim_spec)
        expected_dim_spec = {"type": "default", "dimension": "dim", "outputName": "out"}
        expected = {
            "type": "listFiltered",
            "delegate": expected_dim_spec,
            "values": ["val1", "val2"],
        }

        assert actual == expected

    def test_list_filtered_spec_whitelist(self):
        dim_spec = DimensionSpec("dim", "out").build()
        list_filtered_spec = ListFilteredSpec(["val1", "val2"], is_whitelist=False)
        actual = list_filtered_spec.build(dim_spec)
        expected_dim_spec = {"type": "default", "dimension": "dim", "outputName": "out"}
        expected = {
            "type": "listFiltered",
            "delegate": expected_dim_spec,
            "values": ["val1", "val2"],
            "isWhitelist": False,
        }

        assert actual == expected


class TestRegexFilteredSpec(object):
    def test_regex_filtered_spec(self):
        dim_spec = DimensionSpec("dim", "out").build()
        regex_filtered_spec = RegexFilteredSpec(r"\w+")
        actual = regex_filtered_spec.build(dim_spec)
        expected_dim_spec = {"type": "default", "dimension": "dim", "outputName": "out"}
        expected = {
            "type": "regexFiltered",
            "delegate": expected_dim_spec,
            "pattern": "\\w+",
        }

        assert actual == expected


class TestRegexExtraction(object):
    def test_regex(self):
        ext_fn = RegexExtraction(r"\w+")
        actual = ext_fn.build()
        expected = {"type": "regex", "expr": "\\w+"}

        assert actual == expected


class TestPartialExtraction(object):
    def test_regex(self):
        ext_fn = PartialExtraction(r"\w+")
        actual = ext_fn.build()
        expected = {"type": "partial", "expr": "\\w+"}

        assert actual == expected


class TestJavascriptExtraction(object):
    def test_js_injective(self):
        js_func = "function(x) {return x};"
        ext_fn = JavascriptExtraction(js_func, injective=True)
        actual = ext_fn.build()
        expected = {"type": "javascript", "function": js_func, "injective": True}

        assert actual == expected

    def test_js_not_injective(self):
        js_func = "function(x) {return x};"
        ext_fn = JavascriptExtraction(js_func)
        actual = ext_fn.build()
        expected = {"type": "javascript", "function": js_func, "injective": False}

        assert actual == expected


class TestTimeFormatExtraction(object):
    def test_time_format_all_set(self):
        ext_fn = TimeFormatExtraction("EEEE", "en-US", "Europe/Berlin")
        actual = ext_fn.build()
        expected = {
            "type": "timeFormat",
            "format": "EEEE",
            "locale": "en-US",
            "timeZone": "Europe/Berlin",
        }

        assert actual == expected

    def test_time_format_no_timezone(self):
        ext_fn = TimeFormatExtraction("EEEE", "en-US")
        actual = ext_fn.build()
        expected = {"type": "timeFormat", "format": "EEEE", "locale": "en-US"}

        assert actual == expected

    def test_time_format_only_format(self):
        ext_fn = TimeFormatExtraction("EEEE")
        actual = ext_fn.build()
        expected = {"type": "timeFormat", "format": "EEEE"}

        assert actual == expected


class TestMapLookupExtraction(object):

    mapping = {"foo1": "bar1", "foo2": "bar2"}

    def test_map_default(self):
        ext_fn = MapLookupExtraction(self.mapping)
        actual = ext_fn.build()
        expected = {
            "type": "lookup",
            "lookup": {"type": "map", "map": self.mapping},
            "retainMissingValue": False,
            "replaceMissingValueWith": None,
            "injective": False,
        }

        assert actual == expected

    def test_map_retain_missing(self):
        ext_fn = MapLookupExtraction(self.mapping, retain_missing_values=True)
        actual = ext_fn.build()
        expected = {
            "type": "lookup",
            "lookup": {"type": "map", "map": self.mapping},
            "retainMissingValue": True,
            "replaceMissingValueWith": None,
            "injective": False,
        }

        assert actual == expected

    def test_map_replace_missing(self):
        ext_fn = MapLookupExtraction(self.mapping, replace_missing_values="replacer")
        actual = ext_fn.build()
        expected = {
            "type": "lookup",
            "lookup": {"type": "map", "map": self.mapping},
            "retainMissingValue": False,
            "replaceMissingValueWith": "replacer",
            "injective": False,
        }

        assert actual == expected

    def test_map_injective(self):
        ext_fn = MapLookupExtraction(self.mapping, injective=True)
        actual = ext_fn.build()
        expected = {
            "type": "lookup",
            "lookup": {"type": "map", "map": self.mapping},
            "retainMissingValue": False,
            "replaceMissingValueWith": None,
            "injective": True,
        }

        assert actual == expected


class TestNamespaceLookupExtraction(object):
    def test_map_default(self):
        ext_fn = NamespaceLookupExtraction("foo_namespace")
        actual = ext_fn.build()
        expected = {
            "type": "lookup",
            "lookup": {"type": "namespace", "namespace": "foo_namespace"},
            "retainMissingValue": False,
            "replaceMissingValueWith": None,
            "injective": False,
        }

        assert actual == expected

    def test_map_retain_missing(self):
        ext_fn = NamespaceLookupExtraction("foo_namespace", retain_missing_values=True)
        actual = ext_fn.build()
        expected = {
            "type": "lookup",
            "lookup": {"type": "namespace", "namespace": "foo_namespace"},
            "retainMissingValue": True,
            "replaceMissingValueWith": None,
            "injective": False,
        }

        assert actual == expected

    def test_map_replace_missing(self):
        ext_fn = NamespaceLookupExtraction(
            "foo_namespace", replace_missing_values="replacer"
        )
        actual = ext_fn.build()
        expected = {
            "type": "lookup",
            "lookup": {"type": "namespace", "namespace": "foo_namespace"},
            "retainMissingValue": False,
            "replaceMissingValueWith": "replacer",
            "injective": False,
        }

        assert actual == expected

    def test_map_injective(self):
        ext_fn = NamespaceLookupExtraction("foo_namespace", injective=True)
        actual = ext_fn.build()
        expected = {
            "type": "lookup",
            "lookup": {"type": "namespace", "namespace": "foo_namespace"},
            "retainMissingValue": False,
            "replaceMissingValueWith": None,
            "injective": True,
        }

        assert actual == expected


class TestRegisteredLookupExtraction(object):
    def test_map_default(self):
        ext_fn = RegisteredLookupExtraction("foo_namespace")
        actual = ext_fn.build()
        expected = {
            "type": "registeredLookup",
            "lookup": "foo_namespace",
            "retainMissingValue": False,
            "replaceMissingValueWith": None,
            "injective": False,
        }

        assert actual == expected

    def test_map_retain_missing(self):
        ext_fn = RegisteredLookupExtraction("foo_namespace", retain_missing_values=True)
        actual = ext_fn.build()
        expected = {
            "type": "registeredLookup",
            "lookup": "foo_namespace",
            "retainMissingValue": True,
            "replaceMissingValueWith": None,
            "injective": False,
        }

        assert actual == expected

    def test_map_replace_missing(self):
        ext_fn = RegisteredLookupExtraction(
            "foo_namespace", replace_missing_values="replacer"
        )
        actual = ext_fn.build()
        expected = {
            "type": "registeredLookup",
            "lookup": "foo_namespace",
            "retainMissingValue": False,
            "replaceMissingValueWith": "replacer",
            "injective": False,
        }

        assert actual == expected

    def test_map_injective(self):
        ext_fn = RegisteredLookupExtraction("foo_namespace", injective=True)
        actual = ext_fn.build()
        expected = {
            "type": "registeredLookup",
            "lookup": "foo_namespace",
            "retainMissingValue": False,
            "replaceMissingValueWith": None,
            "injective": True,
        }

        assert actual == expected



# -*- coding: utf-8 -*-

import unittest

from pydruid.db.api import rows_from_chunks


class RowsFromChunksTestSuite(unittest.TestCase):
    def test_rows_from_chunks_empty(self):
        chunks = []
        expected = []
        result = list(rows_from_chunks(chunks))
        self.assertEqual(result, expected)

    def test_rows_from_chunks_single_chunk(self):
        chunks = ['[{"name": "alice"}, {"name": "bob"}, {"name": "charlie"}]']
        expected = [{"name": "alice"}, {"name": "bob"}, {"name": "charlie"}]
        result = list(rows_from_chunks(chunks))
        self.assertEqual(result, expected)

    def test_rows_from_chunks_multiple_chunks(self):
        chunks = ['[{"name": "alice"}, {"name": "b', 'ob"}, {"name": "charlie"}]']
        expected = [{"name": "alice"}, {"name": "bob"}, {"name": "charlie"}]
        result = list(rows_from_chunks(chunks))
        self.assertEqual(result, expected)

    def test_rows_from_chunks_bracket_in_string(self):
        chunks = ['[{"name": "ali{ce"}, {"name": "bob"}]']
        expected = [{"name": "ali{ce"}, {"name": "bob"}]
        result = list(rows_from_chunks(chunks))
        self.assertEqual(result, expected)

    def test_rows_from_chunks_quote_in_string(self):
        chunks = [r'[{"name": "ali\"ce"}, {"name": "bob"}]']
        expected = [{"name": 'ali"ce'}, {"name": "bob"}]
        result = list(rows_from_chunks(chunks))
        self.assertEqual(result, expected)


if __name__ == "__main__":
    unittest.main()



# -*- coding: UTF-8 -*-

import os

from pydruid.utils import query_utils


def open_file(file_path):
    return open(file_path, "w", newline="", encoding="utf-8")


def line_ending():
    return os.linesep


class TestUnicodeWriter:
    def test_writerow(self, tmpdir):
        file_path = tmpdir.join("out.tsv")
        f = open_file(str(file_path))
        w = query_utils.UnicodeWriter(f)
        w.writerow(["value1", "㬓"])
        f.close()
        assert file_path.read() == "value1\t㬓" + line_ending()

    def test_writerows(self, tmpdir):
        file_path = tmpdir.join("out.tsv")
        f = open_file(str(file_path))
        w = query_utils.UnicodeWriter(f)
        w.writerows([["header1", "header2"], ["value1", "㬓"]])
        f.close()
        assert (
            file_path.read()
            == "header1\theader2" + line_ending() + "value1\t㬓" + line_ending()
        )



# -*- coding: UTF-8 -*-

import pytest

from pydruid.utils import dimensions, filters


class TestDimension:
    def test_dimension(self):
        d = filters.Dimension("dim")
        actual = filters.Filter.build_filter(d == "val")
        expected = {"type": "selector", "dimension": "dim", "value": "val"}
        assert actual == expected

    def test_ne_dimension(self):
        d = filters.Dimension("dim")
        actual = filters.Filter.build_filter(d != "val")
        expected = {
            "field": {"dimension": "dim", "type": "selector", "value": "val"},
            "type": "not",
        }
        assert actual == expected


class TestFilter:
    def test_selector_filter(self):
        actual = filters.Filter.build_filter(
            filters.Filter(dimension="dim", value="val")
        )
        expected = {"type": "selector", "dimension": "dim", "value": "val"}
        assert actual == expected

    def test_selector_filter_extraction_fn(self):
        extraction_fn = dimensions.RegexExtraction("([a-b])")
        f = filters.Filter(
            dimension="dim", value="v", extraction_function=extraction_fn
        )
        actual = filters.Filter.build_filter(f)
        expected = {
            "type": "selector",
            "dimension": "dim",
            "value": "v",
            "extractionFn": {"type": "regex", "expr": "([a-b])"},
        }
        assert actual == expected

    def test_extraction_filter(self):
        extraction_fn = dimensions.PartialExtraction("([a-b])")
        f = filters.Filter(
            type="extraction",
            dimension="dim",
            value="v",
            extraction_function=extraction_fn,
        )
        actual = filters.Filter.build_filter(f)
        expected = {
            "type": "extraction",
            "dimension": "dim",
            "value": "v",
            "extractionFn": {"type": "partial", "expr": "([a-b])"},
        }
        assert actual == expected

    def test_javascript_filter(self):
        actual = filters.Filter.build_filter(
            filters.Filter(
                type="javascript", dimension="dim", function="function(x){return true}"
            )
        )
        expected = {
            "type": "javascript",
            "dimension": "dim",
            "function": "function(x){return true}",
        }
        assert actual == expected

    def test_bound_filter(self):
        actual = filters.Filter.build_filter(
            filters.Bound(
                dimension="dim",
                lower="1",
                lowerStrict=True,
                upper="10",
                upperStrict=True,
                ordering="numeric",
            )
        )
        expected = {
            "type": "bound",
            "dimension": "dim",
            "lower": "1",
            "lowerStrict": True,
            "upper": "10",
            "upperStrict": True,
            "alphaNumeric": False,
            "ordering": "numeric",
        }
        assert actual == expected

    def test_bound_filter_with_extraction_function(self):
        f = filters.Bound(
            dimension="d",
            lower="1",
            upper="3",
            upperStrict=True,
            extraction_function=dimensions.RegexExtraction(".*([0-9]+)"),
        )
        actual = filters.Filter.build_filter(f)
        expected = {
            "type": "bound",
            "dimension": "d",
            "lower": "1",
            "lowerStrict": False,
            "upper": "3",
            "upperStrict": True,
            "ordering": "lexicographic",
            "alphaNumeric": False,
            "extractionFn": {"type": "regex", "expr": ".*([0-9]+)"},
        }
        assert actual == expected

    def test_bound_filter_alphanumeric(self):
        actual = filters.Filter.build_filter(
            filters.Bound(
                dimension="dim",
                lower="1",
                lowerStrict=True,
                upper="10",
                upperStrict=True,
                alphaNumeric=True,
            )
        )
        expected = {
            "type": "bound",
            "dimension": "dim",
            "lower": "1",
            "lowerStrict": True,
            "upper": "10",
            "upperStrict": True,
            "alphaNumeric": True,
            "ordering": "lexicographic",
        }
        assert actual == expected

    def test_bound_filter_lower_not_included(self):
        actual = filters.Filter.build_filter(
            filters.Bound(dimension="dim", upper="10", upperStrict=True)
        )
        expected = {
            "type": "bound",
            "dimension": "dim",
            "lower": None,
            "lowerStrict": False,
            "upper": "10",
            "upperStrict": True,
            "alphaNumeric": False,
            "ordering": "lexicographic",
        }
        assert actual == expected

    def test_spatial_filter_rectangle(self):
        actual = filters.Filter.build_filter(
            filters.Spatial(
                dimension="dim",
                bound_type="rectangle",
                minCoords=[100.0, 100.0],
                maxCoords=[100.1, 100.1],
                radius=10.0,
            )
        )
        expected = {
            "type": "spatial",
            "dimension": "dim",
            "bound": {
                "type": "rectangle",
                "minCoords": [100.0, 100.0],
                "maxCoords": [100.1, 100.1],
            },
        }
        assert actual == expected

    def test_spatial_filter_radius(self):
        actual = filters.Filter.build_filter(
            filters.Spatial(
                dimension="dim",
                bound_type="radius",
                coords=[100.0, 100.0],
                radius=100.0,
            )
        )
        expected = {
            "type": "spatial",
            "dimension": "dim",
            "bound": {"type": "radius", "coords": [100.0, 100.0], "radius": 100.0},
        }
        assert actual == expected

    def test_spatial_filter_polygon(self):
        actual = filters.Filter.build_filter(
            filters.Spatial(
                dimension="dim",
                bound_type="polygon",
                abscissa=[2.0, 3.0, 7.0, 8.0],
                ordinate=[4.0, 9.0, 8.0, 1.0],
            )
        )
        expected = {
            "type": "spatial",
            "dimension": "dim",
            "bound": {
                "type": "polygon",
                "abscissa": [2.0, 3.0, 7.0, 8.0],
                "ordinate": [4.0, 9.0, 8.0, 1.0],
            },
        }
        assert actual == expected

    def test_interval_filter(self):
        actual = filters.Filter.build_filter(
            filters.Interval(
                dimension="dim",
                intervals=["2014-10-01T00:00:00.000Z/2014-10-07T00:00:00.000Z"],
            )
        )
        expected = {
            "type": "interval",
            "dimension": "dim",
            "intervals": ["2014-10-01T00:00:00.000Z/2014-10-07T00:00:00.000Z"],
        }
        assert actual == expected

    def test_interval_with_extraction_function(self):
        f = filters.Interval(
            dimension="dim",
            intervals=["2014-10-01T00:00:00.000Z/2014-10-07T00:00:00.000Z"],
            extraction_function=dimensions.RegexExtraction(".*([0-9]+)"),
        )
        actual = filters.Filter.build_filter(f)
        expected = {
            "type": "interval",
            "dimension": "dim",
            "intervals": ["2014-10-01T00:00:00.000Z/2014-10-07T00:00:00.000Z"],
            "extractionFn": {"type": "regex", "expr": ".*([0-9]+)"},
        }
        assert actual == expected

    def test_and_filter(self):
        f1 = filters.Filter(dimension="dim1", value="val1")
        f2 = filters.Filter(dimension="dim2", value="val2")
        actual = filters.Filter.build_filter(f1 & f2)
        expected = {
            "type": "and",
            "fields": [
                {"type": "selector", "dimension": "dim1", "value": "val1"},
                {"type": "selector", "dimension": "dim2", "value": "val2"},
            ],
        }
        assert actual == expected

    def test_and_filter_multiple(self):
        f1 = filters.Filter(dimension="dim1", value="val1")
        f2 = filters.Filter(dimension="dim2", value="val2")
        f3 = filters.Filter(dimension="dim3", value="val3")
        filter = filters.Filter(type="and", fields=[f1, f2, f3])
        actual = filters.Filter.build_filter(filter)
        expected = {
            "type": "and",
            "fields": [
                {"type": "selector", "dimension": "dim1", "value": "val1"},
                {"type": "selector", "dimension": "dim2", "value": "val2"},
                {"type": "selector", "dimension": "dim3", "value": "val3"},
            ],
        }
        assert actual == expected

    def test_or_filter(self):
        f1 = filters.Filter(dimension="dim1", value="val1")
        f2 = filters.Filter(dimension="dim2", value="val2")
        actual = filters.Filter.build_filter(f1 | f2)
        expected = {
            "type": "or",
            "fields": [
                {"type": "selector", "dimension": "dim1", "value": "val1"},
                {"type": "selector", "dimension": "dim2", "value": "val2"},
            ],
        }
        assert actual == expected

    def test_nested_mix_filter(self):
        f1 = filters.Filter(dimension="dim1", value="val1")
        f2 = filters.Filter(dimension="dim2", value="val2")
        f3 = filters.Filter(dimension="dim3", value="val3")
        f4 = filters.Filter(dimension="dim4", value="val4")
        f5 = filters.Filter(dimension="dim5", value="val5")
        f6 = filters.Filter(dimension="dim6", value="val6")
        f7 = filters.Filter(dimension="dim7", value="val7")
        f8 = filters.Filter(dimension="dim8", value="val8")
        actual = filters.Filter.build_filter(
            f1 & ~f2 & f3 & (f4 | ~f5 | f6 | (f7 & ~f8))
        )
        expected = {
            "fields": [
                {"dimension": "dim1", "type": "selector", "value": "val1"},
                {
                    "field": {"dimension": "dim2", "type": "selector", "value": "val2"},
                    "type": "not",
                },
                {"dimension": "dim3", "type": "selector", "value": "val3"},
                {
                    "fields": [
                        {"dimension": "dim4", "type": "selector", "value": "val4"},
                        {
                            "field": {
                                "dimension": "dim5",
                                "type": "selector",
                                "value": "val5",
                            },
                            "type": "not",
                        },
                        {"dimension": "dim6", "type": "selector", "value": "val6"},
                        {
                            "fields": [
                                {
                                    "dimension": "dim7",
                                    "type": "selector",
                                    "value": "val7",
                                },
                                {
                                    "field": {
                                        "dimension": "dim8",
                                        "type": "selector",
                                        "value": "val8",
                                    },
                                    "type": "not",
                                },
                            ],
                            "type": "and",
                        },
                    ],
                    "type": "or",
                },
            ],
            "type": "and",
        }
        assert actual == expected

    def test_or_filter_multiple(self):
        f1 = filters.Filter(dimension="dim1", value="val1")
        f2 = filters.Filter(dimension="dim2", value="val2")
        f3 = filters.Filter(dimension="dim3", value="val3")
        filter = filters.Filter(type="or", fields=[f1, f2, f3])
        actual = filters.Filter.build_filter(filter)
        expected = {
            "type": "or",
            "fields": [
                {"type": "selector", "dimension": "dim1", "value": "val1"},
                {"type": "selector", "dimension": "dim2", "value": "val2"},
                {"type": "selector", "dimension": "dim3", "value": "val3"},
            ],
        }
        assert actual == expected

    def test_not_filter(self):
        f = ~filters.Filter(dimension="dim", value="val")
        actual = filters.Filter.build_filter(f)
        # Call `build_filter` twice to make sure it does not
        # change the passed filter object argument `f`.
        actual = filters.Filter.build_filter(f)
        expected = {
            "type": "not",
            "field": {"type": "selector", "dimension": "dim", "value": "val"},
        }
        assert actual == expected

    def test_nested_not_or_filter(self):
        f1 = filters.Filter(dimension="dim1", value="val1")
        f2 = filters.Filter(dimension="dim2", value="val2")
        actual = filters.Filter.build_filter(~(f1 | f2))
        expected = {
            "type": "not",
            "field": {
                "type": "or",
                "fields": [
                    {"type": "selector", "dimension": "dim1", "value": "val1"},
                    {"type": "selector", "dimension": "dim2", "value": "val2"},
                ],
            },
        }
        assert actual == expected

    def test_in_filter(self):
        actual = filters.Filter.build_filter(
            filters.Filter(type="in", dimension="dim", values=["val1", "val2", "val3"])
        )
        expected = {
            "type": "in",
            "dimension": "dim",
            "values": ["val1", "val2", "val3"],
        }
        assert actual == expected

    def test_not_in_filter(self):
        actual = filters.Filter.build_filter(
            ~filters.Filter(type="in", dimension="dim", values=["val1", "val2", "val3"])
        )
        expected = {
            "type": "not",
            "field": {
                "type": "in",
                "dimension": "dim",
                "values": ["val1", "val2", "val3"],
            },
        }
        assert actual == expected

    def test_invalid_filter(self):
        with pytest.raises(NotImplementedError):
            filters.Filter(type="invalid", dimension="dim", value="val")

    def test_columnComparison_filter(self):
        actual = filters.Filter.build_filter(
            filters.Filter(
                type="columnComparison",
                dimensions=["dim1", dimensions.DimensionSpec("dim2", "dim2")],
            )
        )
        expected = {
            "type": "columnComparison",
            "dimensions": [
                "dim1",
                {"type": "default", "dimension": "dim2", "outputName": "dim2"},
            ],
        }
        assert actual == expected

    def test_search_filter(self):
        # Without caseSensitive param - default:false
        actual = filters.Filter.build_filter(
            filters.Filter(type="search", dimension="dim", value="val")
        )
        expected = {
            "type": "search",
            "dimension": "dim",
            "query": {"type": "contains", "caseSensitive": "false", "value": "val"},
        }
        assert actual == expected

        # With caseSensitive param
        actual = filters.Filter.build_filter(
            filters.Filter(
                type="search", dimension="dim", value="val", caseSensitive="true"
            )
        )
        expected = {
            "type": "search",
            "dimension": "dim",
            "query": {"type": "contains", "caseSensitive": "true", "value": "val"},
        }
        assert actual == expected

    def test_like_filter(self):
        actual = filters.Filter.build_filter(
            filters.Filter(type="like", dimension="dim", pattern="%val%")
        )
        expected = {"type": "like", "dimension": "dim", "pattern": "%val%"}
        assert actual == expected



# -*- coding: UTF-8 -*-
import textwrap
import urllib
from io import StringIO
from unittest.mock import Mock, patch

import pytest

from pydruid.client import PyDruid
from pydruid.query import Query
from pydruid.utils.aggregators import doublesum
from pydruid.utils.filters import Dimension


def create_client():
    return PyDruid("http://localhost:8083", "druid/v2/")


def create_blank_query():
    return Query({}, "none")


def _http_error(code, msg, data=""):
    # Need a file-like object for the response data
    fp = StringIO(data)
    return urllib.error.HTTPError(
        url="http://fakeurl:8080/druid/v2/", hdrs={}, code=code, msg=msg, fp=fp
    )


class TestPyDruid:
    @patch("pydruid.client.urllib.request.urlopen")
    def test_druid_returns_error(self, mock_urlopen):
        # given
        mock_urlopen.side_effect = _http_error(500, "Druid error")
        client = create_client()

        # when / then
        with pytest.raises(IOError):
            client.topn(
                datasource="testdatasource",
                granularity="all",
                intervals="2015-12-29/pt1h",
                aggregations={"count": doublesum("count")},
                dimension="user_name",
                metric="count",
                filter=Dimension("user_lang") == "en",
                threshold=1,
                context={"timeout": 1000},
            )

    @patch("pydruid.client.urllib.request.urlopen")
    def test_druid_returns_html_error(self, mock_urlopen):
        # given
        message = textwrap.dedent(
            """
            <html>
            <head>
            <meta http-equiv="Content-Type" content="text/html;charset=ISO-8859-1"/>
            <title>Error 500 </title>
            </head>
            <body>
            <h2>HTTP ERROR: 500</h2>
            <p>Problem accessing /druid/v2/. Reason:
            <pre>    javax.servlet.ServletException: java.lang.OutOfMemoryError: GC overhead limit exceeded</pre></p>
            <hr /><a href="http://eclipse.org/jetty">Powered by Jetty:// 9.3.19.v20170502</a><hr/>
            </body>
            </html>
        """
        ).strip()
        mock_urlopen.side_effect = _http_error(500, "Internal Server Error", message)
        client = create_client()

        # when / then
        with pytest.raises(IOError) as e:
            client.topn(
                datasource="testdatasource",
                granularity="all",
                intervals="2015-12-29/pt1h",
                aggregations={"count": doublesum("count")},
                dimension="user_name",
                metric="count",
                filter=Dimension("user_lang") == "en",
                threshold=1,
                context={"timeout": 1000},
            )

        assert (
            str(e.value)
            == textwrap.dedent(
                """
            HTTP Error 500: Internal Server Error 
             Druid Error: javax.servlet.ServletException: java.lang.OutOfMemoryError: GC overhead limit exceeded 
             Query is: {
                "aggregations": [
                    {
                        "fieldName": "count",
                        "name": "count",
                        "type": "doubleSum"
                    }
                ],
                "context": {
                    "timeout": 1000
                },
                "dataSource": "testdatasource",
                "dimension": "user_name",
                "filter": {
                    "dimension": "user_lang",
                    "type": "selector",
                    "value": "en"
                },
                "granularity": "all",
                "intervals": "2015-12-29/pt1h",
                "metric": "count",
                "queryType": "topN",
                "threshold": 1
            }
        """
            ).strip()
        )

    @patch("pydruid.client.urllib.request.urlopen")
    def test_druid_returns_results(self, mock_urlopen):
        # given
        response = Mock()
        response.read.return_value = """
            [ {
  "timestamp" : "2015-12-30T14:14:49.000Z",
  "result" : [ {
    "dimension" : "aaaa",
    "metric" : 100
  } ]
            } ]
        """.encode(
            "utf-8"
        )
        mock_urlopen.return_value = response
        client = create_client()

        # when
        top = client.topn(
            datasource="testdatasource",
            granularity="all",
            intervals="2015-12-29/pt1h",
            aggregations={"count": doublesum("count")},
            dimension="user_name",
            metric="count",
            filter=Dimension("user_lang") == "en",
            threshold=1,
            context={"timeout": 1000},
        )

        # then
        assert top is not None
        assert len(top.result) == 1
        assert len(top.result[0]["result"]) == 1

    @patch("pydruid.client.urllib.request.urlopen")
    def test_client_allows_to_export_last_query(self, mock_urlopen):
        # given
        response = Mock()
        response.read.return_value = """
            [ {
  "timestamp" : "2015-12-30T14:14:49.000Z",
  "result" : [ {
    "dimension" : "aaaa",
    "metric" : 100
  } ]
            } ]
        """.encode(
            "utf-8"
        )
        mock_urlopen.return_value = response
        client = create_client()
        client.topn(
            datasource="testdatasource",
            granularity="all",
            intervals="2015-12-29/pt1h",
            aggregations={"count": doublesum("count")},
            dimension="user_name",
            metric="count",
            filter=Dimension("user_lang") == "en",
            threshold=1,
            context={"timeout": 1000},
        )

        # when / then
        # assert that last_query.export_tsv method was called (it should throw an exception, given empty path)
        with pytest.raises(TypeError):
            client.export_tsv(None)

    @patch("pydruid.client.urllib.request.urlopen")
    def test_client_auth_creds(self, mock_urlopen):
        client = create_client()
        query = create_blank_query()
        client.set_basic_auth_credentials("myUsername", "myPassword")
        headers, _, _ = client._prepare_url_headers_and_body(query)
        assert headers["Authorization"] == "Basic bXlVc2VybmFtZTpteVBhc3N3b3Jk"



# -*- coding: utf-8 -*-

import unittest
from collections import namedtuple
from io import BytesIO
from unittest.mock import patch

from requests.models import Response

from pydruid.db.api import apply_parameters, Cursor


class CursorTestSuite(unittest.TestCase):
    @patch("requests.post")
    def test_execute(self, requests_post_mock):
        response = Response()
        response.status_code = 200
        response.raw = BytesIO(
            b'[{"name": "alice"}, {"name": "bob"}, {"name": "charlie"}]'
        )
        requests_post_mock.return_value = response
        Row = namedtuple("Row", ["name"])

        cursor = Cursor("http://example.com/")
        cursor.execute("SELECT * FROM table")
        result = cursor.fetchall()
        expected = [Row(name="alice"), Row(name="bob"), Row(name="charlie")]
        self.assertEqual(result, expected)

    @patch("requests.post")
    def test_execute_empty_result(self, requests_post_mock):
        response = Response()
        response.status_code = 200
        response.raw = BytesIO(b"[]")
        requests_post_mock.return_value = response

        cursor = Cursor("http://example.com/")
        cursor.execute("SELECT * FROM table")
        result = cursor.fetchall()
        expected = []
        self.assertEqual(result, expected)

    @patch("requests.post")
    def test_context(self, requests_post_mock):
        response = Response()
        response.status_code = 200
        response.raw = BytesIO(b"[]")
        requests_post_mock.return_value = response

        url = "http://example.com/"
        query = "SELECT * FROM table"
        context = {"source": "unittest"}

        cursor = Cursor(url, user=None, password=None, context=context)
        cursor.execute(query)

        requests_post_mock.assert_called_with(
            "http://example.com/",
            auth=None,
            stream=True,
            headers={"Content-Type": "application/json"},
            json={"query": query, "context": context, "header": False},
            verify=True,
            cert=None,
            proxies=None,
        )

    @patch("requests.post")
    def test_header_false(self, requests_post_mock):
        response = Response()
        response.status_code = 200
        response.raw = BytesIO(b'[{"name": "alice"}]')
        requests_post_mock.return_value = response
        Row = namedtuple("Row", ["name"])

        url = "http://example.com/"
        query = "SELECT * FROM table"

        cursor = Cursor(url, header=False)
        cursor.execute(query)
        result = cursor.fetchall()
        self.assertEqual(result, [Row(name="alice")])

        self.assertEqual(
            cursor.description, [("name", 1, None, None, None, None, True)]
        )

    @patch("requests.post")
    def test_header_true(self, requests_post_mock):
        response = Response()
        response.status_code = 200
        response.raw = BytesIO(b'[{"name": null}, {"name": "alice"}]')
        requests_post_mock.return_value = response
        Row = namedtuple("Row", ["name"])

        url = "http://example.com/"
        query = "SELECT * FROM table"

        cursor = Cursor(url, header=True)
        cursor.execute(query)
        result = cursor.fetchall()
        self.assertEqual(result, [Row(name="alice")])
        self.assertEqual(cursor.description, [("name", None)])

    @patch("requests.post")
    def test_names_with_underscores(self, requests_post_mock):
        response = Response()
        response.status_code = 200
        response.raw = BytesIO(b'[{"_name": null}, {"_name": "alice"}]')
        requests_post_mock.return_value = response
        Row = namedtuple("Row", ["_name"], rename=True)

        url = "http://example.com/"
        query = "SELECT * FROM table"

        cursor = Cursor(url, header=True)
        cursor.execute(query)
        result = cursor.fetchall()
        self.assertEqual(result, [Row(_0="alice")])
        self.assertEqual(cursor.description, [("_name", None)])

    def test_apply_parameters(self):
        self.assertEqual(
            apply_parameters('SELECT 100 AS "100%"', None), 'SELECT 100 AS "100%"'
        )

        self.assertEqual(
            apply_parameters('SELECT 100 AS "100%"', {}), 'SELECT 100 AS "100%"'
        )

        self.assertEqual(
            apply_parameters('SELECT %(key)s AS "100%%"', {"key": 100}),
            'SELECT 100 AS "100%"',
        )

        self.assertEqual(apply_parameters("SELECT %(key)s", {"key": "*"}), "SELECT *")

        self.assertEqual(
            apply_parameters("SELECT %(key)s", {"key": "bar"}), "SELECT 'bar'"
        )

        self.assertEqual(
            apply_parameters("SELECT %(key)s", {"key": True}), "SELECT TRUE"
        )

        self.assertEqual(
            apply_parameters("SELECT %(key)s", {"key": False}), "SELECT FALSE"
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: UTF-8 -*-
#
# Copyright 2016 Metamarkets Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest.mock import Mock

import pytest
import tornado
import tornado.ioloop
import tornado.web
from tornado.httpclient import HTTPError
from tornado.testing import AsyncHTTPTestCase

from pydruid.async_client import AsyncPyDruid
from pydruid.utils.aggregators import doublesum
from pydruid.utils.filters import Dimension


class FailureHandler(tornado.web.RequestHandler):
    def post(self):
        raise HTTPError(500, "Druid error", response="Druid error")


class SuccessHandler(tornado.web.RequestHandler):
    def post(self):
        self.write(
            """
            [ {
  "timestamp" : "2015-12-30T14:14:49.000Z",
  "result" : [ {
    "dimension" : "aaaa",
    "metric" : 100
  } ]
            } ]
            """
        )


class TestAsyncPyDruid(AsyncHTTPTestCase):
    def get_app(self):
        return tornado.web.Application(
            [
                (r"/druid/v2/fail_request", FailureHandler),
                (r"/druid/v2/return_results", SuccessHandler),
            ]
        )

    @tornado.testing.gen_test
    def test_druid_returns_error(self):
        # given
        client = AsyncPyDruid(
            "http://localhost:%s" % (self.get_http_port(),), "druid/v2/fail_request"
        )

        # when / then
        with pytest.raises(IOError):
            yield client.topn(
                datasource="testdatasource",
                granularity="all",
                intervals="2015-12-29/pt1h",
                aggregations={"count": doublesum("count")},
                dimension="user_name",
                metric="count",
                filter=Dimension("user_lang") == "en",
                threshold=1,
                context={"timeout": 1000},
            )

    @tornado.testing.gen_test
    def test_druid_returns_results(self):
        # given
        client = AsyncPyDruid(
            "http://localhost:%s" % (self.get_http_port(),), "druid/v2/return_results"
        )

        # when
        top = yield client.topn(
            datasource="testdatasource",
            granularity="all",
            intervals="2015-12-29/pt1h",
            aggregations={"count": doublesum("count")},
            dimension="user_name",
            metric="count",
            filter=Dimension("user_lang") == "en",
            threshold=1,
            context={"timeout": 1000},
        )

        # then
        self.assertIsNotNone(top)
        self.assertEqual(len(top.result), 1)
        self.assertEqual(len(top.result[0]["result"]), 1)

    @tornado.testing.gen_test
    def test_client_allows_to_export_last_query(self):
        # given
        client = AsyncPyDruid(
            "http://localhost:%s" % (self.get_http_port(),), "druid/v2/return_results"
        )
        yield client.topn(
            datasource="testdatasource",
            granularity="all",
            intervals="2015-12-29/pt1h",
            aggregations={"count": doublesum("count")},
            dimension="user_name",
            metric="count",
            filter=Dimension("user_lang") == "en",
            threshold=1,
            context={"timeout": 1000},
        )

        # when / then
        # assert that last_query.export_tsv method was called (it should throw an exception, given empty path)
        with pytest.raises(TypeError):
            client.export_tsv(None)

    @tornado.testing.gen_test
    def test_client_allows_passing_default_parameters(self):
        # given
        client = AsyncPyDruid(
            "http://localhost:%s" % (self.get_http_port(),),
            "druid/v2/return_results",
            defaults=dict(request_timeout=120),
        )
        top = yield client.topn(
            datasource="testdatasource",
            granularity="all",
            intervals="2015-12-29/pt1h",
            aggregations={"count": doublesum("count")},
            dimension="user_name",
            metric="count",
            filter=Dimension("user_lang") == "en",
            threshold=1,
            context={"timeout": 1000},
        )

        # then
        self.assertIsNotNone(top)
        self.assertEqual(len(top.result), 1)
        self.assertEqual(len(top.result[0]["result"]), 1)

    @tornado.testing.gen_test
    def test_client_allows_passing_http_client(self):
        # given
        client = AsyncPyDruid(
            "http://localhost:%s" % (self.get_http_port(),),
            "druid/v2/return_results",
            http_client="tornado.curl_httpclient.CurlAsyncHTTPClient",
        )
        top = yield client.topn(
            datasource="testdatasource",
            granularity="all",
            intervals="2015-12-29/pt1h",
            aggregations={"count": doublesum("count")},
            dimension="user_name",
            metric="count",
            filter=Dimension("user_lang") == "en",
            threshold=1,
            context={"timeout": 1000},
        )

        # then
        self.assertIsNotNone(top)
        self.assertEqual(len(top.result), 1)
        self.assertEqual(len(top.result[0]["result"]), 1)



def build_dimension(dim):
    if isinstance(dim, DimensionSpec):
        dim = dim.build()

    return dim


class DimensionSpec(object):
    def __init__(
        self, dimension, output_name, extraction_function=None, filter_spec=None
    ):
        self._dimension = dimension
        self._output_name = output_name
        self._extraction_function = extraction_function
        self._filter_spec = filter_spec

    def build(self):
        dimension_spec = {
            "type": "default",
            "dimension": self._dimension,
            "outputName": self._output_name,
        }

        if self._extraction_function is not None:
            dimension_spec["type"] = "extraction"
            dimension_spec["extractionFn"] = self._extraction_function.build()

        if self._filter_spec is not None:
            dimension_spec = self._filter_spec.build(dimension_spec)

        return dimension_spec


class FilteredSpec(object):

    filter_type = None

    def build(self, delegate):
        dimension_spec = {"type": self.filter_type, "delegate": delegate}
        return dimension_spec


class ListFilteredSpec(FilteredSpec):

    filter_type = "listFiltered"

    def __init__(self, values, is_whitelist=True):
        self._values = values
        self._is_whitelist = is_whitelist

    def build(self, dimension_spec):
        filtered_dimension_spec = super(ListFilteredSpec, self).build(dimension_spec)
        filtered_dimension_spec["values"] = self._values

        if not self._is_whitelist:
            filtered_dimension_spec["isWhitelist"] = False

        return filtered_dimension_spec


class RegexFilteredSpec(FilteredSpec):

    filter_type = "regexFiltered"

    def __init__(self, pattern):
        self._pattern = pattern

    def build(self, dimension_spec):
        filtered_dimension_spec = super(RegexFilteredSpec, self).build(dimension_spec)
        filtered_dimension_spec["pattern"] = self._pattern

        return filtered_dimension_spec


class ExtractionFunction(object):

    extraction_type = None

    def build(self):
        return {"type": self.extraction_type}


class BaseRegexExtraction(ExtractionFunction):
    def __init__(self, expr):
        super(BaseRegexExtraction, self).__init__()
        self._expr = expr

    def build(self):
        extractor = super(BaseRegexExtraction, self).build()
        extractor["expr"] = self._expr

        return extractor


class RegexExtraction(BaseRegexExtraction):

    extraction_type = "regex"


class PartialExtraction(BaseRegexExtraction):

    extraction_type = "partial"


class JavascriptExtraction(ExtractionFunction):

    extraction_type = "javascript"

    def __init__(self, func, injective=False):
        super(JavascriptExtraction, self).__init__()
        self._func = func
        self._injective = injective

    def build(self):
        extractor = super(JavascriptExtraction, self).build()
        extractor["function"] = self._func
        extractor["injective"] = self._injective

        return extractor


class TimeFormatExtraction(ExtractionFunction):

    extraction_type = "timeFormat"

    def __init__(self, format, locale=None, time_zone=None):
        super(TimeFormatExtraction, self).__init__()
        self._format = format
        self._locale = locale
        self._time_zone = time_zone

    def build(self):
        extractor = super(TimeFormatExtraction, self).build()
        extractor["format"] = self._format
        if self._locale:
            extractor["locale"] = self._locale
        if self._time_zone:
            extractor["timeZone"] = self._time_zone

        return extractor


class LookupExtraction(ExtractionFunction):

    extraction_type = "lookup"
    lookup_type = None

    def __init__(
        self, retain_missing_values=False, replace_missing_values=None, injective=False
    ):
        super(LookupExtraction, self).__init__()
        self._retain_missing_values = retain_missing_values
        self._replace_missing_values = replace_missing_values
        self._injective = injective

    def build(self):
        extractor = super(LookupExtraction, self).build()
        extractor["lookup"] = self.build_lookup()
        extractor["retainMissingValue"] = self._retain_missing_values
        extractor["replaceMissingValueWith"] = self._replace_missing_values
        extractor["injective"] = self._injective

        return extractor

    def build_lookup(self):
        return {"type": self.lookup_type}


class MapLookupExtraction(LookupExtraction):

    lookup_type = "map"

    def __init__(self, mapping, **kwargs):
        super(MapLookupExtraction, self).__init__(**kwargs)
        self._mapping = mapping

    def build_lookup(self):
        lookup = super(MapLookupExtraction, self).build_lookup()
        lookup["map"] = self._mapping

        return lookup


class NamespaceLookupExtraction(LookupExtraction):

    lookup_type = "namespace"

    def __init__(self, namespace, **kwargs):
        super(NamespaceLookupExtraction, self).__init__(**kwargs)
        self._namespace = namespace

    def build_lookup(self):
        lookup = super(NamespaceLookupExtraction, self).build_lookup()
        lookup["namespace"] = self._namespace

        return lookup


class RegisteredLookupExtraction(LookupExtraction):

    extraction_type = "registeredLookup"

    def __init__(self, reglookup, **kwargs):
        super(RegisteredLookupExtraction, self).__init__(**kwargs)
        self._lookup = reglookup

    def build_lookup(self):
        return self._lookup



# -*- coding: UTF-8 -*-
#
# Copyright 2016 Metamarkets Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import csv
import os

import pandas
import pytest
from pandas.testing import assert_frame_equal

from pydruid.query import Query, QueryBuilder
from pydruid.utils import aggregators, filters, having, postaggregator


def create_query_with_results():
    query = Query({}, "timeseries")
    query.result = [
        {
            "result": {"value1": 1, "value2": "㬓"},
            "timestamp": "2015-01-01T00:00:00.000-05:00",
        },
        {
            "result": {"value1": 2, "value2": "㬓"},
            "timestamp": "2015-01-02T00:00:00.000-05:00",
        },
    ]
    return query


EXPECTED_RESULTS_PANDAS = [
    {"timestamp": "2015-01-01T00:00:00.000-05:00", "value1": 1, "value2": "㬓"},
    {"timestamp": "2015-01-02T00:00:00.000-05:00", "value1": 2, "value2": "㬓"},
]


def expected_results_csv_reader():
    # csv.DictReader does not perform promotion to int64
    expected_results = []
    for element in EXPECTED_RESULTS_PANDAS:
        modified_elem = element.copy()
        modified_elem.update({"value1": str(modified_elem["value1"])})
        expected_results.append(modified_elem)
    return expected_results


class TestQueryBuilder:
    def test_build_query(self):
        # given
        expected_query_dict = {
            "queryType": None,
            "dataSource": "things",
            "aggregations": [{"fieldName": "thing", "name": "count", "type": "count"}],
            "postAggregations": [
                {
                    "fields": [
                        {"fieldName": "sum", "type": "fieldAccess"},
                        {"fieldName": "count", "type": "fieldAccess"},
                    ],
                    "fn": "/",
                    "name": "avg",
                    "type": "arithmetic",
                }
            ],
            "pagingSpec": {"pagingIdentifies": {}, "threshold": 1},
            "filter": {"dimension": "one", "type": "selector", "value": 1},
            "having": {"aggregation": "sum", "type": "greaterThan", "value": 1},
            "new_key": "value",
        }

        builder = QueryBuilder()

        # when
        query = builder.build_query(
            None,
            {
                "datasource": "things",
                "aggregations": {"count": aggregators.count("thing")},
                "post_aggregations": {
                    "avg": (postaggregator.Field("sum") / postaggregator.Field("count"))
                },
                "paging_spec": {"pagingIdentifies": {}, "threshold": 1},
                "filter": filters.Dimension("one") == 1,
                "having": having.Aggregation("sum") > 1,
                "new_key": "value",
            },
        )

        # then
        assert query.query_dict == expected_query_dict

    def test_build_query_none_type(self):
        # given
        expected_query_dict = {
            "queryType": None,
            "dataSource": "things",
            "aggregations": [{"fieldName": "thing", "name": "count", "type": "count"}],
            "filter": {"dimension": "one", "type": "selector", "value": 1},
            "having": {"aggregation": "sum", "type": "greaterThan", "value": 1},
            "dimension": "dim1",
        }

        builder = QueryBuilder()

        # when
        builder_dict = {
            "datasource": "things",
            "aggregations": {"count": aggregators.count("thing")},
            "filter": filters.Dimension("one") == 1,
            "having": having.Aggregation("sum") > 1,
            "dimension": "dim1",
        }
        query = builder.build_query(None, builder_dict)

        # then
        assert query.query_dict == expected_query_dict

        # you should be able to pass `None` to dimension/having/filter
        for v in ["dimension", "having", "filter"]:
            expected_query_dict[v] = None
            builder_dict[v] = None

            query = builder.build_query(None, builder_dict)

            assert query.query_dict == expected_query_dict

    def test_validate_query(self):
        # given
        builder = QueryBuilder()

        # when
        builder.validate_query(None, ["validkey"], {"validkey": "value"})

        # then
        pytest.raises(
            ValueError,
            builder.validate_query,
            *[None, ["validkey"], {"invalidkey": "value"}]
        )

    def test_union_datasource(self):
        # Given
        expected_query_dict = {"queryType": None, "dataSource": "things"}
        builder = QueryBuilder()
        # when
        builder_dict = {"datasource": "things"}
        query = builder.build_query(None, builder_dict)
        # then
        assert query.query_dict == expected_query_dict

        # Given
        expected_query_dict = {
            "queryType": None,
            "dataSource": {
                "type": "union",
                "dataSources": ["things", "others", "more"],
            },
        }
        builder = QueryBuilder()
        # when
        builder_dict = {"datasource": ["things", "others", "more"]}
        query = builder.build_query(None, builder_dict)
        # then
        assert query.query_dict == expected_query_dict

        # Given check that it rejects non-string items
        builder = QueryBuilder()
        builder_dict = {"datasource": ["things", 123]}
        with pytest.raises(ValueError):
            query = builder.build_query(None, builder_dict)

    def test_build_subquery(self):
        # given
        expected_query_dict = {
            "query": {
                "queryType": "groupBy",
                "dataSource": "things",
                "aggregations": [
                    {"fieldName": "thing", "name": "count", "type": "count"}
                ],
                "postAggregations": [
                    {
                        "fields": [
                            {"fieldName": "sum", "type": "fieldAccess"},
                            {"fieldName": "count", "type": "fieldAccess"},
                        ],
                        "fn": "/",
                        "name": "avg",
                        "type": "arithmetic",
                    }
                ],
                "filter": {"dimension": "one", "type": "selector", "value": 1},
                "having": {"aggregation": "sum", "type": "greaterThan", "value": 1},
            },
            "type": "query",
        }

        builder = QueryBuilder()

        # when
        subquery_dict = builder.subquery(
            {
                "datasource": "things",
                "aggregations": {"count": aggregators.count("thing")},
                "post_aggregations": {
                    "avg": (postaggregator.Field("sum") / postaggregator.Field("count"))
                },
                "filter": filters.Dimension("one") == 1,
                "having": having.Aggregation("sum") > 1,
            }
        )

        # then
        assert subquery_dict == expected_query_dict


class TestQuery:
    def test_export_tsv(self, tmpdir):
        query = create_query_with_results()
        file_path = tmpdir.join("out.tsv")
        query.export_tsv(str(file_path))

        with open(str(file_path)) as tsv_file:
            reader = csv.DictReader(tsv_file, delimiter="\t")
            actual = [line for line in reader]
            assert actual == expected_results_csv_reader()

    def test_export_pandas(self):
        query = create_query_with_results()
        df = query.export_pandas()
        expected_df = pandas.DataFrame(EXPECTED_RESULTS_PANDAS)
        assert_frame_equal(df, expected_df, check_like=True)

        query = Query({}, "timeseries")
        df = query.export_pandas()
        assert_frame_equal(df, pandas.DataFrame())

    def test_query_acts_as_a_wrapper_for_raw_result(self):
        # given
        query = create_query_with_results()

        # then
        assert len(query) == 2
        assert isinstance(query[0], dict)
        assert isinstance(query[1], dict)



# -*- coding: UTF-8 -*-

from copy import deepcopy
from operator import itemgetter

from pydruid.utils import aggregators, filters


class TestAggregators:
    def test_aggregators(self):
        aggs = [
            ("longsum", "longSum"),
            ("longmin", "longMin"),
            ("longmax", "longMax"),
            ("doublesum", "doubleSum"),
            ("doublemin", "doubleMin"),
            ("doublemax", "doubleMax"),
            ("count", "count"),
            ("hyperunique", "hyperUnique"),
            ("stringfirst", "stringFirst"),
            ("stringlast", "stringLast"),
        ]
        aggs_funcs = [
            (getattr(aggregators, agg_name), agg_type) for agg_name, agg_type in aggs
        ]
        for f, agg_type in aggs_funcs:
            assert f("metric") == {"type": agg_type, "fieldName": "metric"}

    def test_filtered_aggregator(self):
        filter_ = filters.Filter(dimension="dim", value="val")
        aggs = [
            aggregators.count("metric1"),
            aggregators.longsum("metric2"),
            aggregators.doublesum("metric3"),
            aggregators.doublemin("metric4"),
            aggregators.doublemax("metric5"),
            aggregators.hyperunique("metric6"),
            aggregators.cardinality("dim1"),
            aggregators.cardinality(["dim1", "dim2"], by_row=True),
            aggregators.thetasketch("dim1"),
            aggregators.thetasketch("metric7"),
            aggregators.thetasketch("metric8", isinputthetasketch=True, size=8192),
        ]
        for agg in aggs:
            expected = {
                "type": "filtered",
                "filter": {"type": "selector", "dimension": "dim", "value": "val"},
                "aggregator": agg,
            }
            actual = aggregators.filtered(filter_, agg)
            assert actual == expected

    def test_nested_filtered_aggregator(self):
        filter1 = filters.Filter(dimension="dim1", value="val")
        filter2 = filters.Filter(dimension="dim2", value="val")
        agg = aggregators.filtered(
            filter1, aggregators.filtered(filter2, aggregators.count("metric1"))
        )
        actual = aggregators.build_aggregators({"agg_name": agg})
        # the innermost aggregation must have 'agg_name'
        expected = [
            {
                "type": "filtered",
                "aggregator": {
                    "type": "filtered",
                    "aggregator": {
                        "fieldName": "metric1",
                        "type": "count",
                        "name": "agg_name",
                    },
                    "filter": {"dimension": "dim2", "value": "val", "type": "selector"},
                },
                "filter": {"dimension": "dim1", "value": "val", "type": "selector"},
            }
        ]
        assert expected == actual

    def test_build_aggregators(self):
        agg_input = {
            "agg1": aggregators.count("metric1"),
            "agg2": aggregators.longsum("metric2"),
            "agg3": aggregators.doublesum("metric3"),
            "agg4": aggregators.doublemin("metric4"),
            "agg5": aggregators.doublemax("metric5"),
            "agg6": aggregators.hyperunique("metric6"),
            "agg7": aggregators.cardinality("dim1"),
            "agg8": aggregators.cardinality(["dim1", "dim2"], by_row=True),
            "agg9": aggregators.thetasketch("dim1"),
            "agg10": aggregators.thetasketch("metric7"),
            "agg11": aggregators.thetasketch(
                "metric8", isinputthetasketch=True, size=8192
            ),
        }
        built_agg = aggregators.build_aggregators(agg_input)
        expected = [
            {"name": "agg1", "type": "count", "fieldName": "metric1"},
            {"name": "agg2", "type": "longSum", "fieldName": "metric2"},
            {"name": "agg3", "type": "doubleSum", "fieldName": "metric3"},
            {"name": "agg4", "type": "doubleMin", "fieldName": "metric4"},
            {"name": "agg5", "type": "doubleMax", "fieldName": "metric5"},
            {"name": "agg6", "type": "hyperUnique", "fieldName": "metric6"},
            {
                "name": "agg7",
                "type": "cardinality",
                "fieldNames": ["dim1"],
                "byRow": False,
            },
            {
                "name": "agg8",
                "type": "cardinality",
                "fieldNames": ["dim1", "dim2"],
                "byRow": True,
            },
            {
                "name": "agg9",
                "type": "thetaSketch",
                "fieldName": "dim1",
                "isInputThetaSketch": False,
                "size": 16384,
            },
            {
                "name": "agg10",
                "type": "thetaSketch",
                "fieldName": "metric7",
                "isInputThetaSketch": False,
                "size": 16384,
            },
            {
                "name": "agg11",
                "type": "thetaSketch",
                "fieldName": "metric8",
                "isInputThetaSketch": True,
                "size": 8192,
            },
        ]
        assert sorted(built_agg, key=itemgetter("name")) == sorted(
            expected, key=itemgetter("name")
        )

    def test_build_filtered_aggregator(self):
        filter_ = filters.Filter(dimension="dim", value="val")
        agg_input = {
            "agg1": aggregators.filtered(filter_, aggregators.count("metric1")),
            "agg2": aggregators.filtered(filter_, aggregators.longsum("metric2")),
            "agg3": aggregators.filtered(filter_, aggregators.doublesum("metric3")),
            "agg4": aggregators.filtered(filter_, aggregators.doublemin("metric4")),
            "agg5": aggregators.filtered(filter_, aggregators.doublemax("metric5")),
            "agg6": aggregators.filtered(filter_, aggregators.hyperunique("metric6")),
            "agg7": aggregators.filtered(filter_, aggregators.cardinality("dim1")),
            "agg8": aggregators.filtered(
                filter_, aggregators.cardinality(["dim1", "dim2"], by_row=True)
            ),
            "agg9": aggregators.filtered(filter_, aggregators.thetasketch("dim1")),
            "agg10": aggregators.filtered(filter_, aggregators.thetasketch("metric7")),
            "agg11": aggregators.filtered(
                filter_,
                aggregators.thetasketch("metric8", isinputthetasketch=True, size=8192),
            ),
        }
        base = {
            "type": "filtered",
            "filter": {"type": "selector", "dimension": "dim", "value": "val"},
        }

        aggs = [
            {"name": "agg1", "type": "count", "fieldName": "metric1"},
            {"name": "agg2", "type": "longSum", "fieldName": "metric2"},
            {"name": "agg3", "type": "doubleSum", "fieldName": "metric3"},
            {"name": "agg4", "type": "doubleMin", "fieldName": "metric4"},
            {"name": "agg5", "type": "doubleMax", "fieldName": "metric5"},
            {"name": "agg6", "type": "hyperUnique", "fieldName": "metric6"},
            {
                "name": "agg7",
                "type": "cardinality",
                "fieldNames": ["dim1"],
                "byRow": False,
            },
            {
                "name": "agg8",
                "type": "cardinality",
                "fieldNames": ["dim1", "dim2"],
                "byRow": True,
            },
            {
                "name": "agg9",
                "type": "thetaSketch",
                "fieldName": "dim1",
                "isInputThetaSketch": False,
                "size": 16384,
            },
            {
                "name": "agg10",
                "type": "thetaSketch",
                "fieldName": "metric7",
                "isInputThetaSketch": False,
                "size": 16384,
            },
            {
                "name": "agg11",
                "type": "thetaSketch",
                "fieldName": "metric8",
                "isInputThetaSketch": True,
                "size": 8192,
            },
        ]
        expected = []
        for agg in aggs:
            exp = deepcopy(base)
            exp.update({"aggregator": agg})
            expected.append(exp)

        built_agg = aggregators.build_aggregators(agg_input)
        expected = sorted(
            built_agg, key=lambda k: itemgetter("name")(itemgetter("aggregator")(k))
        )
        actual = sorted(
            expected, key=lambda k: itemgetter("name")(itemgetter("aggregator")(k))
        )
        assert expected == actual



# -*- coding: UTF-8 -*-

import pytest

from pydruid.utils.filters import Filter
from pydruid.utils.having import Aggregation, Dimension, Having


class TestHaving:
    def test_equalTo_having(self):
        h1 = Having(type="equalTo", aggregation="sum", value=1)
        actual = Having.build_having(h1)
        expected = {"type": "equalTo", "aggregation": "sum", "value": 1}
        assert actual == expected

    def test_equalTo_having_Aggregation(self):
        h1 = Aggregation("sum") == 1
        actual = Having.build_having(h1)
        expected = {"type": "equalTo", "aggregation": "sum", "value": 1}
        assert actual == expected

    def test_lessThan_having(self):
        h1 = Having(type="lessThan", aggregation="sum", value=1)
        actual = Having.build_having(h1)
        expected = {"type": "lessThan", "aggregation": "sum", "value": 1}
        assert actual == expected

    def test_lessThan_having_Aggregation(self):
        h1 = Aggregation("sum") < 1
        actual = Having.build_having(h1)
        expected = {"type": "lessThan", "aggregation": "sum", "value": 1}
        assert actual == expected

    def test_greaterThan_having(self):
        h1 = Having(type="greaterThan", aggregation="sum", value=1)
        actual = Having.build_having(h1)
        expected = {"type": "greaterThan", "aggregation": "sum", "value": 1}
        assert actual == expected

    def test_greaterThan_having_Aggregation(self):
        h1 = Aggregation("sum") > 1
        actual = Having.build_having(h1)
        expected = {"type": "greaterThan", "aggregation": "sum", "value": 1}
        assert actual == expected

    def test_and_having(self):
        h1 = Aggregation("sum1") > 1
        h2 = Aggregation("sum2") > 2
        actual = Having.build_having(h1 & h2)
        expected = {
            "type": "and",
            "havingSpecs": [
                {"type": "greaterThan", "aggregation": "sum1", "value": 1},
                {"type": "greaterThan", "aggregation": "sum2", "value": 2},
            ],
        }
        assert actual == expected

    def test_or_having(self):
        h1 = Aggregation("sum1") > 1
        h2 = Aggregation("sum2") > 2
        actual = Having.build_having(h1 | h2)
        expected = {
            "type": "or",
            "havingSpecs": [
                {"type": "greaterThan", "aggregation": "sum1", "value": 1},
                {"type": "greaterThan", "aggregation": "sum2", "value": 2},
            ],
        }
        assert actual == expected

    def test_not_having(self):
        h1 = Aggregation("sum") > 1
        actual = Having.build_having(~h1)
        expected = {
            "type": "not",
            "havingSpec": {"type": "greaterThan", "aggregation": "sum", "value": 1},
        }
        assert actual == expected

    def test_dimSelector_having(self):
        h1 = Having(type="dimSelector", dimension="foo", value="bar")
        actual = Having.build_having(h1)
        expected = {"type": "dimSelector", "dimension": "foo", "value": "bar"}
        assert actual == expected

    def test_dimension(self):
        h1 = Dimension("foo") == "bar"
        actual = Having.build_having(h1)
        expected = {"type": "dimSelector", "dimension": "foo", "value": "bar"}
        assert actual == expected

    def test_query_filter_having(self):
        f1 = Filter(type="selector", dimension="foo", value="bar")
        query_filter = Filter.build_filter(f1)
        h1 = Having(type="filter", filter=query_filter)
        actual = Having.build_having(h1)
        expected = {
            "type": "filter",
            "filter": {"type": "selector", "dimension": "foo", "value": "bar"},
        }
        assert actual == expected

    def test_not_exists_having_type(self):
        with pytest.raises(NotImplementedError):
            Having(type="notExists")



#
# Copyright 2013 Metamarkets Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
try:
    import simplejson as json
except ImportError:
    import json

from .dimensions import build_dimension


class Filter:

    # filter types supporting extraction function
    _FILTERS_WITH_EXTR_FN = (
        "selector",
        "regex",
        "javascript",
        "in",
        "bound",
        "interval",
        "extraction",
    )

    def __init__(self, extraction_function=None, ordering="lexicographic", **args):

        type_ = args.get("type", "selector")

        if extraction_function is not None:
            if type_ not in self._FILTERS_WITH_EXTR_FN:
                raise ValueError(
                    "Filter of type {0} doesn't support "
                    "extraction function".format(type_)
                )
        elif type_ == "extraction":
            raise ValueError(
                "Filter of type extraction requires extraction " "function"
            )

        self.extraction_function = extraction_function

        self.filter = {"filter": {"type": type_}}

        if type_ == "selector":
            self.filter["filter"].update(
                {"dimension": args["dimension"], "value": args["value"]}
            )
        elif type_ == "javascript":
            self.filter["filter"].update(
                {"dimension": args["dimension"], "function": args["function"]}
            )
        elif type_ == "and":
            self.filter["filter"].update({"fields": args["fields"]})
        elif type_ == "or":
            self.filter["filter"].update({"fields": args["fields"]})
        elif type_ == "not":
            self.filter["filter"].update({"field": args["field"]})
        elif type_ == "in":
            self.filter["filter"].update(
                {"dimension": args["dimension"], "values": args["values"]}
            )
        elif type_ == "regex":
            self.filter["filter"].update(
                {"dimension": args["dimension"], "pattern": args["pattern"]}
            )
        elif type_ == "bound":
            self.filter["filter"].update(
                {
                    "dimension": args["dimension"],
                    "lower": args["lower"],
                    "lowerStrict": args["lowerStrict"],
                    "upper": args["upper"],
                    "upperStrict": args["upperStrict"],
                    "alphaNumeric": args["alphaNumeric"],
                    "ordering": ordering,
                }
            )
        elif type_ == "columnComparison":
            self.filter["filter"].update({"dimensions": args["dimensions"]})
        elif type_ == "interval":
            self.filter["filter"].update(
                {"dimension": args["dimension"], "intervals": args["intervals"]}
            )
        elif type_ == "extraction":
            self.filter["filter"].update(
                {"dimension": args["dimension"], "value": args["value"]}
            )
        elif type_ == "search":
            self.filter["filter"].update(
                {
                    "dimension": args["dimension"],
                    "query": {
                        "type": "contains",
                        "value": args["value"],
                        "caseSensitive": args.get("caseSensitive", "false"),
                    },
                }
            )
        elif type_ == "like":
            self.filter["filter"].update(
                {"dimension": args["dimension"], "pattern": args["pattern"]}
            )
        elif type_ == "spatial":
            self.filter["filter"].update(
                {"dimension": args["dimension"], "bound": args["bound"]}
            )
        else:
            raise NotImplementedError("Filter type: {0} does not exist".format(type_))

    def show(self):
        print(json.dumps(self.filter, indent=4))

    def __and__(self, x):
        if self.filter["filter"]["type"] == "and":
            # if `self` is already `and`, don't create a new filter
            # but just append `x` to the filter fields.
            self.filter["filter"]["fields"].append(x)
            return self
        return Filter(type="and", fields=[self, x])

    def __or__(self, x):
        if self.filter["filter"]["type"] == "or":
            # if `self` is already `or`, don't create a new filter
            # but just append `x` to the filter fields.
            self.filter["filter"]["fields"].append(x)
            return self
        return Filter(type="or", fields=[self, x])

    def __invert__(self):
        return Filter(type="not", field=self)

    @staticmethod
    def build_filter(filter_obj):
        filter = filter_obj.filter["filter"]
        if filter["type"] in ["and", "or"]:
            filter = filter.copy()  # make a copy so we don't overwrite `fields`
            filter["fields"] = [Filter.build_filter(f) for f in filter["fields"]]
        elif filter["type"] in ["not"]:
            filter = filter.copy()
            filter["field"] = Filter.build_filter(filter["field"])
        elif filter["type"] in ["columnComparison"]:
            filter = filter.copy()
            filter["dimensions"] = [build_dimension(d) for d in filter["dimensions"]]

        if filter_obj.extraction_function is not None:
            if filter is filter_obj.filter["filter"]:  # copy if not yet copied
                filter = filter.copy()
            filter["extractionFn"] = filter_obj.extraction_function.build()

        return filter


class Dimension:
    def __init__(self, dim):
        self.dimension = dim

    def __eq__(self, other):
        return Filter(dimension=self.dimension, value=other)

    def __ne__(self, other):
        return ~Filter(dimension=self.dimension, value=other)


class JavaScript:
    def __init__(self, dim):
        self.dimension = dim

    def __eq__(self, func):
        return Filter(type="javascript", dimension=self.dimension, function=func)


class Bound(Filter):
    """
    Bound filter can be used to filter by comparing dimension values to an
    upper value or/and a lower value.

    :ivar str dimension: Dimension to filter on.
    :ivar str lower: Lower bound.
    :ivar str upper: Upper bound.
    :ivar bool lowerStrict: Strict lower inclusion. Initial value: False
    :ivar bool upperStrict: Strict upper inclusion. Initial value: False
    :ivar bool alphaNumeric: Numeric comparison. Initial value: False
        NOTE: For backwards compatibility - Use "ordering" instead.
    :ivar str ordering: Sorting Order. Initial value: lexicographic
        Specifies the sorting order to use when comparing values against the bound.
        Can be one of the following values: "lexicographic", "alphanumeric", "numeric",
        "strlen", "version". See Sorting Orders
        https://druid.apache.org/docs/latest/querying/filters.html#bound-filter
        for more details.
    :ivar ExtractionFunction extraction_function: extraction function to use,
                                                  if not None
    """

    def __init__(
        self,
        dimension,
        lower=None,
        upper=None,
        lowerStrict=False,
        upperStrict=False,
        alphaNumeric=False,
        ordering="lexicographic",
        extraction_function=None,
    ):
        if not lower and not upper:
            raise ValueError("Must include either lower or upper or both")
        Filter.__init__(
            self,
            type="bound",
            dimension=dimension,
            lower=lower,
            upper=upper,
            lowerStrict=lowerStrict,
            upperStrict=upperStrict,
            alphaNumeric=alphaNumeric,
            ordering=ordering,
            extraction_function=extraction_function,
        )


class Interval(Filter):
    """
    Interval filter can be used to filter by comparing dimension(__time)
    values to a list of intervals.

    :ivar str dimension: Dimension to filter on.
    :ivar list intervals: List of ISO-8601 intervals of data to filter out.
    :ivar ExtractionFunction extraction_function: extraction function to use,
                                                  if not None
    """

    def __init__(self, dimension, intervals, extraction_function=None):

        Filter.__init__(
            self,
            type="interval",
            dimension=dimension,
            intervals=intervals,
            extraction_function=extraction_function,
        )


class Spatial(Filter):
    """
    Spatial filter can be used to filter by spatial bounds

    :ivar str dimension: Dimension to filter on.
    :ivar str bound_type: Spatial bound type: ['rectangle','radius','polygon'].
    :param `**kwargs`: addition arguments required for the selected bound type:
        'rectange': 'minCoords' and 'maxCoords'
        'radius': 'coords' and 'radius'
        'polygon': 'abscissa' and 'ordinate'
    """

    def __init__(self, dimension, bound_type, **args):

        _bound = {"type": bound_type}

        if bound_type == "rectangle":
            if not args["minCoords"] or not args["maxCoords"]:
                raise ValueError(
                    "Rectangle bound must include both minCoords and maxCoords"
                )
            _bound["minCoords"] = args["minCoords"]
            _bound["maxCoords"] = args["maxCoords"]
        elif bound_type == "radius":
            if not args["coords"] or not args["radius"]:
                raise ValueError("Radius bound must include both coords and radius")
            _bound["coords"] = args["coords"]
            _bound["radius"] = args["radius"]
        elif bound_type == "polygon":
            if not args["abscissa"] or not args["ordinate"]:
                raise ValueError(
                    "Polygon bound must include both abscissa and ordinate"
                )
            _bound["abscissa"] = args["abscissa"]
            _bound["ordinate"] = args["ordinate"]
        else:
            raise ValueError("Unsupport Spatial Bound type: {0}".format(bound_type))

        Filter.__init__(self, type="spatial", dimension=dimension, bound=_bound)
